# region Imports
from __future__ import annotations
from typing import Tuple, overload
from enum import Enum

import uno
from com.sun.star.chart import XChartDocument
from com.sun.star.container import XNameAccess
from com.sun.star.sheet import XSpreadsheet
from com.sun.star.table import XTableChart
from com.sun.star.table import XTableCharts
from com.sun.star.table import XTableChartsSupplier
from com.sun.star.document import XEmbeddedObjectSupplier
from com.sun.star.lang import XMultiServiceFactory
from com.sun.star.chart import XDiagram
from com.sun.star.drawing import XDrawPageSupplier
from com.sun.star.drawing import XShape
from com.sun.star.drawing import XDrawPage
from com.sun.star.beans import XPropertySet
from com.sun.star.text import XTextDocument
from com.sun.star.text import XTextContent
from com.sun.star.chart import XAxisXSupplier
from com.sun.star.chart import XAxisYSupplier
from com.sun.star.chart import XTwoAxisYSupplier
from com.sun.star.chart import XChartDataArray

from ..utils.kind.chart_diagram_kind import ChartDiagramKind as ChartDiagramKind
from ..utils.kind.drawing_shape_kind import DrawingShapeKind as DrawingShapeKind

from ooo.dyn.table.cell_range_address import CellRangeAddress as CellRangeAddress
from ooo.dyn.awt.rectangle import Rectangle
from ooo.dyn.text.vert_orientation import VertOrientation
from ooo.dyn.text.hori_orientation import HoriOrientation
from ooo.dyn.chart.chart_data_caption import ChartDataCaptionEnum as ChartDataCaptionEnum
from ooo.dyn.chart.data_label_placement import DataLabelPlacementEnum as DataLabelPlacementEnum
from ooo.dyn.chart.chart_solid_type import ChartSolidTypeEnum as ChartSolidTypeEnum
from ooo.dyn.chart.chart_symbol_type import ChartSymbolType as ChartSymbolType
from ooo.dyn.chart.chart_regression_curve_type import ChartRegressionCurveType as ChartRegressionCurveType

from ..utils import lo as mLo
from ..utils import props as mProps
from . import draw as mDraw
from ..utils import info as mInfo
from ..exceptions import ex as mEx
from ..utils.data_type.intensity import Intensity

# endregion Imports
class Chart:
    """Chart Class"""

    # region insert_chart()
    @overload
    @classmethod
    def insert_chart(
        cls, *, slide: XDrawPage, x: int, y: int, width: int, height: int, diagram_name: ChartDiagramKind | str
    ) -> XChartDocument:
        ...

    @overload
    @classmethod
    def insert_chart(
        cls, *, doc: XTextDocument, x: int, y: int, width: int, height: int, diagram_name: ChartDiagramKind | str
    ) -> XChartDocument:
        ...

    @overload
    @classmethod
    def insert_chart(
        cls,
        *,
        sheet: XSpreadsheet,
        chart_name: str,
        cells_range: CellRangeAddress,
        width: int,
        height: int,
        diagram_name: ChartDiagramKind | str,
    ) -> XChartDocument:
        ...

    @overload
    @classmethod
    def insert_chart(
        cls,
        *,
        sheet: XSpreadsheet,
        chart_name: str,
        cells_range: CellRangeAddress,
        x: int,
        y: int,
        width: int,
        height: int,
        diagram_name: ChartDiagramKind | str,
    ) -> XChartDocument:
        ...

    @classmethod
    def _insert_chart_slide(
        cls, slide: XDrawPage, x: int, y: int, width: int, height: int, diagram_name: ChartDiagramKind | str
    ) -> XChartDocument:
        try:
            ashape = mDraw.Draw.add_shape(
                slide=slide,
                shape_type=DrawingShapeKind.OLE2_SHAPE,
                x=x,
                y=y,
                width=width,
                height=height,
            )
            chart_doc = cls._get_chart_doc_shape(ashape)
            cls.set_chart_type(chart_doc=chart_doc, diagram_name=diagram_name)
            return chart_doc
        except mEx.ChartError:
            raise
        except Exception as e:
            raise mEx.ChartError("Error inserting chart slide") from e

    @classmethod
    def _insert_chart_sheet(
        cls,
        sheet: XSpreadsheet,
        chart_name: str,
        cells_range: CellRangeAddress,
        x: int,
        y: int,
        width: int,
        height: int,
        diagram_name: ChartDiagramKind | str,
    ) -> XChartDocument:
        try:
            tbl_charts = cls.get_table_charts(sheet)
            tc_access = mLo.Lo.qi(XNameAccess, tbl_charts)
            if tc_access is None:
                raise mEx.ChartNoAccessError("Unable to get name access to chart table")

            if tc_access.hasByName(chart_name):
                raise mEx.ChartExistingError(f"A chart table called {chart_name} already exist")

            rect = Rectangle(x * 1000, y * 1000, width * 1000, height * 1000)

            addrs = [cells_range]
            # first boolean: has column headers?; second boolean: has row headers?
            tbl_charts.addNewByName(chart_name, rect, addrs, True, True)
            # 2nd last arg: whether the topmost row of the source data will be used
            # to set labels for the category axis or the legend.
            # last arg: whether the leftmost column of the source data will be
            # used to set labels for the category axis or the legend.

            tbl_chart = cls.get_table_chart(tc_access, chart_name)
            chart_doc = cls.get_chart_doc(tbl_chart)

            if diagram_name is not None:
                cls.set_chart_type(chart_doc, diagram_name)
            return chart_doc
        except mEx.ChartNoAccessError:
            raise
        except mEx.ChartExistingError:
            raise
        except mEx.ChartError:
            raise
        except Exception as e:
            raise mEx.ChartError(f'Error inserting "{chart_name}" into sheet') from e

    @classmethod
    def _insert_chart_doc(
        cls, doc: XTextDocument, x: int, y: int, width: int, height: int, diagram_name: ChartDiagramKind | str
    ) -> XChartDocument:
        try:
            tc = mLo.Lo.create_instance_msf(XTextContent, "com.sun.star.text.TextEmbeddedObject", raise_err=True)
            ps = mLo.Lo.qi(XPropertySet, tc, True)
            ps.setPropertyValue("CLSID", str(mLo.Lo.CLSID.CHART))

            xtext = doc.getText()
            cursor = xtext.createTextCursor()

            # insert embedded object in text -> object will be created
            xtext.insertTextContent(cursor, tc, True)

            # set size and position
            shape = mLo.Lo.qi(XShape, tc, True)
            shape.setSize(mDraw.Draw.Size(width * 1_000, height * 1_000))

            ps.setPropertyValue("VertOrient", VertOrientation.NONE)
            ps.setPropertyValue("HoriOrient", HoriOrientation.NONE)
            ps.setPropertyValue("VertOrientPosition", y * 1_000)
            ps.setPropertyValue("HoriOrientPosition", x * 1_000)

            # retrieve the chart document as model of the OLE shape
            chart_doc = mLo.Lo.qi(XChartDocument, ps.getPropertyValue("Model"))
            if chart_doc is not None:
                cls.set_chart_type(chart_doc=chart_doc, diagram_name=diagram_name)
        except Exception as e:
            raise mEx.ChartError("Unable to insert chart") from e

    @classmethod
    def insert_chart(cls, *args, **kwargs) -> XChartDocument:
        """
        Insert a chart using the given name as name of the OLE object and
        the range as corresponding range of data to be used for rendering.
        The chart is placed in the sheet for charts at position (1,1)
        extending as large as given in chart size.

        The diagram name must be the name of a diagram service (i.e one
        in ``com.sun.star.chart``) that can be
        instantiated via the factory of the chart document

        Args:
            slide (XDrawPage): Draw page
            doc (XTextDocument): Document
            sheet (XSpreadsheet): Spreadsheet
            chart_name (str): Name of chart
            cells_range (CellRangeAddress): Cells range
            width (int): Width
            height (int): Height
            diagram_name (DiagramKind | str): Diagram Name
            x (int, optional): Chart X position. Defaults to 1.
            y (int, optional): Chart y Position. Defaults to 1.

        Raises:
            ChartNoAccessError: If unable to get access to chart table.
            ChartExistingError: If chart already exist.
            ChartError: If unable to insert chart.

        Returns:
            XChartDocument: Chart Document
        """
        if len(args) > 0:
            raise TypeError(f"Insert_chart() takes 0 positional arguments but {len(args)} was given")

        count = len(kwargs)

        if not count in (6, 8):
            raise TypeError("insert_chart() got an invalid number of arguments")

        if count == 6:
            if "slide" in kwargs:
                # insert_chart( slide: XDrawPage, x: int, y: int, width: int, height: int, diagram_name: ChartDiagramKind | str) -> XChartDocument:
                return cls._insert_chart_slide(**kwargs)
            if "doc" in kwargs:
                # insert_chart(doc: XTextDocument, x: int, y: int, width: int, height: int, diagram_name: ChartDiagramKind | str) -> XChartDocument:
                return cls._insert_chart_doc(**kwargs)
            # else
            # insert_chart(sheet: XSpreadsheet, chart_name: str, cells_range: CellRangeAddress, width: int, height: int, diagram_name: ChartDiagramKind | str,) -> XChartDocument:
            kargs = kwargs.copy()
            kargs["x"] = 1
            kargs["y"] = 1
            return cls._insert_chart_sheet(**kargs)

        return cls._insert_chart_sheet(**kwargs)

    # endregion insert_chart()

    @classmethod
    def get_table_chart_access(cls, sheet: XSpreadsheet) -> XNameAccess:
        """
        Get XnameAccess interface of sheet charts

        Args:
            sheet (XSpreadsheet): Spreadsheet

        Raises:
            ChartError: if error occurs.

        Returns:
            XNameAccess: Named Access interface
        """
        try:
            tbl_charts = cls.get_table_charts(sheet)
            return mLo.Lo.qi(XNameAccess, tbl_charts, True)
        except Exception as e:
            raise mEx.ChartError("Error getting table chart access") from e

    @staticmethod
    def get_table_charts(sheet: XSpreadsheet) -> XTableCharts:
        """
        Gets Table Charts of a sheet

        Args:
            sheet (XSpreadsheet): Spreadsheet

        Raises:
            ChartError: if error occurs.

        Returns:
            XTableCharts: Table Charts interface
        """
        try:
            # get the supplier for the charts
            charts_supplier = mLo.Lo.qi(XTableChartsSupplier, sheet, True)
            return charts_supplier.getCharts()
        except Exception as e:
            raise mEx.ChartError("Error getting table chart access") from e

    @classmethod
    def get_chart_names_list(cls, sheet: XSpreadsheet) -> Tuple[str, ...]:
        """
        Gets Tuple of chart names as strings

        Args:
            sheet (XSpreadsheet): Spreadsheet

        Returns:
            Tuple[str]: Tuple of strings or empty tuple if no charts are found
        """
        try:
            tbl_charts = cls.get_table_charts(sheet)
        except mEx.ChartError:
            return ()
        return tbl_charts.getElementNames()

    @staticmethod
    def get_table_chart(tc_access: XNameAccess, chart_name: str) -> XTableChart:
        """
        Gets a table chart.

        Args:
            tc_access (XNameAccess): Name Access
            chart_name (str): Chart Name

        Raises:
            ChartError: If unable to get Table Chart

        Returns:
            XTableChart: Table Chart Interface.
        """
        try:
            tbl_chart = mLo.Lo.qi(XTableChart, tc_access.getByName(chart_name), True)
            return tbl_chart
        except Exception as e:
            raise mEx.ChartError(f"Could Not Access: {chart_name}") from e

    # region get_chart_doc()

    @staticmethod
    def _get_chart_doc_tbl_chart(tbl_chart: XTableChart) -> XChartDocument:
        """
        Gets chart document

        Args:
            tbl_chart (XTableChart): Table Chart

        Returns:
            XChartDocument: Chart Document Interface
        """
        try:
            eos = mLo.Lo.qi(XEmbeddedObjectSupplier, tbl_chart, True)
            intf = eos.getEmbeddedObject()
            return mLo.Lo.qi(XChartDocument, intf, True)
        except Exception as e:
            raise mEx.ChartError("Unable to get chart.") from e

    @classmethod
    def _get_chart_doc_sheet_chart_name(cls, sheet: XSpreadsheet, chart_name: str) -> XChartDocument:
        tc_access = cls.get_table_chart_access(sheet)

        if not tc_access.hasByName(chart_name):
            raise mEx.ChartNotExistingError(f'No table chart called "{chart_name}" found')

        tbl_chart = cls.get_table_chart(tc_access=tc_access, chart_name=chart_name)
        return cls._get_chart_doc_tbl_chart(tbl_chart)

    @staticmethod
    def _get_chart_doc_shape(shape: XShape) -> XChartDocument:
        try:
            # change the LOL shape into a chart
            try:
                shape_props = mLo.Lo.qi(XPropertySet, shape, True)
            except mEx.MissingInterfaceError as ex:
                raise mEx.ChartError("Unable to access shape properties") from ex
            # set the class id for charts
            # shape_props.setPropertyValue("CLSID", str(mLo.Lo.CLSID.CHART))

            # retrieve the chart document as model of the OLE shape
            chart_doc = mLo.Lo.qi(XChartDocument, shape_props.getPropertyValue("Model"), True)
            return chart_doc
        except mEx.ChartError:
            raise
        except Exception as e:
            raise mEx.ChartError("Couldn't change the OLE shape into a chart.") from e

    @overload
    @classmethod
    def get_chart_doc(cls, shape: XShape) -> XChartDocument:
        ...

    @overload
    @classmethod
    def get_chart_doc(cls, tbl_chart: XTableChart) -> XChartDocument:
        ...

    @overload
    @classmethod
    def get_chart_doc(cls, sheet: XSpreadsheet, chart_name: str) -> XChartDocument:
        ...

    @classmethod
    def get_chart_doc(cls, *args, **kwargs) -> XChartDocument:
        """
        Gets chart document

        Args:
            shape (XShape): Shape object
            tbl_chart (XTableChart): Table Chart
            sheet: SpreadSheet
            chart_name (str): Chart Name

        Raises:
            ChartNotExistingError: If chart is not existing.
            ChartError: If chart error occurs.

        Returns:
            XChartDocument: Chart Document Interface
        """
        ordered_keys = (1, 2)
        kargs_len = len(kwargs)
        count = len(args) + kargs_len

        def get_kwargs() -> dict:
            ka = {}
            if kargs_len == 0:
                return ka
            valid_keys = ("tbl_chart", "shape", "sheet", "chart_name")
            check = all(key in valid_keys for key in kwargs.keys())
            if not check:
                raise TypeError("get_chart_doc() got an unexpected keyword argument")
            keys = ("tbl_chart", "sheet", "shape")
            for key in keys:
                if key in kwargs:
                    ka[1] = kwargs[key]
                    break
            if count == 1:
                return ka
            ka[2] = kwargs.get("chart_name", None)
            return ka

        if not count in (1, 2):
            raise TypeError("get_chart_doc() got an invalid number of arguments")

        kargs = get_kwargs()
        for i, arg in enumerate(args):
            kargs[ordered_keys[i]] = arg

        if count == 1:
            # def get_chart_doc(cls, shape: XShape)
            # def get_chart_doc(cls, tbl_chart: XTableChart)
            shape = mLo.Lo.qi(XShape, kargs[1])
            if shape is None:
                return cls._get_chart_doc_tbl_chart(kargs[1])
            else:
                return cls._get_chart_doc_shape(shape)

        return cls._get_chart_doc_sheet_chart_name(kargs[1], kargs[2])

    # endregion get_chart_doc()

    @staticmethod
    def set_chart_type(chart_doc: XChartDocument, diagram_name: ChartDiagramKind | str) -> None:
        """
        Sets chart diagram name

        Args:
            chart_doc (XChartDocument): Chart Document
            diagram_name (DiagramEnum | str): Diagram Name

        Raises:
            ChartError: If error occurs.
        """
        try:
            msf = mLo.Lo.qi(XMultiServiceFactory, chart_doc, True)
            chart_type = f"com.sun.star.chart.{diagram_name}"
            diagram = mLo.Lo.qi(XDiagram, msf.createInstance(chart_type), True)
            chart_doc.setDiagram(diagram)
        except Exception as e:
            raise mEx.ChartError(f"Could not set the chart type to: {chart_type}") from e

    @staticmethod
    def get_chart_type(chart_doc: XChartDocument) -> str:
        """
        Gets chart document chart type

        Args:
            chart_doc (XChartDocument): Chart Document

        Returns:
            str: Chart Type
        """
        try:
            result = chart_doc.getDiagram().getDiagramType()
            if result is None:
                raise mEx.ChartError("Error, result of get chart type is None")
            return result
        except mEx.ChartError:
            raise
        except Exception as e:
            raise mEx.ChartError("Error getting chart type")

    @classmethod
    def remove_chart(cls, sheet: XSpreadsheet, chart_name: str) -> bool:
        """
        Removes a chart from spreadsheet current charts

        Args:
            sheet (XSpreadsheet): Spreadsheet
            chart_name (str): Chart Name to remove

        Returns:
            bool: ``True`` if chart is found and removed; Otherwise, ``False``
        """
        mLo.Lo.print(f'Removing chart "{chart_name}"')
        try:
            tbl_charts = cls.get_table_charts(sheet)
            tc_access = mLo.Lo.qi(XNameAccess, tbl_charts, True)
        except mEx.ChartError as ce:
            mLo.Lo.print(f'Error get table charts for "{chart_name}"')
            mLo.Lo.print(f"  {ce}")
            return False
        except mEx.MissingInterfaceError as e:
            mLo.Lo.print("Unable to get name access to chart table")
            mLo.Lo.print(f"  {ce}")
            return False

        if tc_access.hasByName(chart_name):
            tbl_charts.removeByName(chart_name)
            mLo.Lo.print(f'Chart table "{chart_name}" removed.')
            return True

        mLo.Lo.print(f'Chart table "{chart_name}" not found.')
        return False

    @staticmethod
    def set_visible(sheet: XSpreadsheet, is_visible: bool) -> None:
        """
        Set visible state of charts.

        Args:
            sheet (XSpreadsheet): Spreadsheet
            is_visible (bool): Visible State. ``True`` shows charts, ``False`` hides charts.

        Raises:
            ChartError: If error occurs.

        Returns:
            None:

        Note:
            Any chart on sheet errors when attempting to set visibility then then it is skipped
            and no error is raised.
        """

        # get draw page supplier for chart sheet
        try:
            page_supplier = mLo.Lo.qi(XDrawPageSupplier, sheet, True)
            draw_page = page_supplier.getDrawPage()
            num_shapes = draw_page.getCount()
            chart_class_id = str(mLo.Lo.CLSID.CHART)
        except Exception as e:
            raise mEx.ChartError("Error setting visibility of sheet") from e

        for i in range(num_shapes):
            try:
                shape = mLo.Lo.qi(XShape, draw_page.getByIndex(i), True)
                class_id = str(mProps.Props.get(shape, "CLSID"))

                if class_id.casefold() == chart_class_id:
                    mProps.Props.set(shape, Visible=is_visible)
            except Exception as e:
                mLo.Lo.print("Error setting visibility of chart")
                mLo.Lo.print(f"  {e}")

    @staticmethod
    def get_chart_shape(sheet: XSpreadsheet) -> XShape:
        """
        Gets the shape of the first found chart

        Args:
            sheet (XSpreadsheet): Spreadsheet

        Raises:
            ShapeMissingError: If unable to find a chart on the sheet.
            ShapeError: If any other error occurs.

        Returns:
            XShape: Chart Shape.
        """
        try:
            # get draw page supplier for chart sheet
            page_supplier = mLo.Lo.qi(XDrawPageSupplier, sheet, True)
            draw_page = page_supplier.getDrawPage()
            num_shapes = draw_page.getCount()
            chart_class_id = str(mLo.Lo.CLSID.CHART)
            shape = None
            for i in range(num_shapes):
                try:
                    shape = mLo.Lo.qi(XShape, draw_page.getByIndex(i), True)
                    class_id = str(mProps.Props.get(shape, "CLSID"))

                    if class_id.casefold() == chart_class_id:
                        break
                except Exception:
                    pass
            if shape is None:
                raise mEx.ShapeMissingError("Unable to find any chart on sheet")
            else:
                mLo.Lo.print("Found a chart")
            return shape
        except mEx.ShapeMissingError:
            raise
        except Exception as e:
            raise mEx.ShapeError("Error getting chart shape") from e

    # region adjust properties
    @staticmethod
    def get_title(chart_doc: XChartDocument) -> str:
        """
        Gets title for a chart document.

        Args:
            chart_doc (XChartDocument): Chart document

        Raises:
            ChartError: If error occurs.

        Returns:
            str: Chart Title
        """
        try:
            title = mProps.Props.get(chart_doc.getTitle(), "String")
            if title is None:
                raise mEx.ChartError("Error getting title for chart document. Property returned None")
            return str(title)
        except mEx.ChartError:
            raise
        except Exception as e:
            raise mEx.ChartError("Error geting chart document title") from e

    @staticmethod
    def set_title(chart_doc: XChartDocument, title: str) -> None:
        """
        Set title for a chart document

        Args:
            chart_doc (XChartDocument): Chart document
            title (str): Title text

        Raises:
            ChartError: If error occurs

        Returns:
            None:
        """
        try:
            shape = chart_doc.getTitle()
            mProps.Props.set(shape, String=title)
        except Exception as e:
            raise mEx.ChartError("Error setting title for chart document.") from e

    @staticmethod
    def get_sub_title(chart_doc: XChartDocument) -> str:
        """
        Gets subtitle for a chart document.

        Args:
            chart_doc (XChartDocument): Chart document

        Raises:
            ChartError: If error occurs.

        Returns:
            str: Chart Subtitle
        """
        try:
            sub_title = mProps.Props.get(chart_doc.getSubTitle(), "String")
            if sub_title is None:
                raise mEx.ChartError("Error getting subtitle for chart document. Property returned None")
            return str(sub_title)
        except mEx.ChartError:
            raise
        except Exception as e:
            raise mEx.ChartError("Error geting chart document subtitle") from e

    @staticmethod
    def set_subtitle(chart_doc: XChartDocument, subtitle: str) -> None:
        """
        Set subtitle for a chart document

        Args:
            chart_doc (XChartDocument): Chart document
            subtitle (str): Subtitle text

        Raises:
            ChartError: If error occurs

        Returns:
            None:
        """
        try:
            # in java this next line was in error and should have been getSubTitle()
            # chartDoc.getTitle();
            shape = chart_doc.getSubTitle()
            mProps.Props.set(shape, String=subtitle)
            # mProps.Props.set_property(shape, "HasSubTitle", True)
        except Exception as e:
            raise mEx.ChartError("Error setting title for chart document.") from e

    @staticmethod
    def set_x_axis_title(chart_doc: XChartDocument, title: str) -> None:
        """
        Sets the X Axis Title

        Args:
            chart_doc (XChartDocument): Chart Document
            title (str): Title text

        Raises:
            DiagramNotExistingError: If ``chart_doc`` diagram does not exist
            ServiceNotSupported: If ``com.sun.star.chart.ChartAxisXSupplier`` service is not supported by diagram.
            ChartError: If any other error occurs.

        Returns:
            None:
        """
        try:
            diagram = chart_doc.getDiagram()
            if diagram is None:
                raise mEx.DiagramNotExistingError("Diagram does not exist")

            if not mInfo.Info.support_service(diagram, "com.sun.star.chart.ChartAxisXSupplier"):
                raise mEx.ServiceNotSupported("com.sun.star.chart.ChartAxisXSupplier")

            mProps.Props.set(diagram, HasXAxisDescription=True)
            axis = mLo.Lo.qi(XAxisXSupplier, diagram, True)
            title_shape = axis.getXAxisTitle()
            mProps.Props.set_property(title_shape, "String", title)
            # mProps.Props.set_property(title_shape, "HasXAxisTitle", True)
        except mEx.DiagramNotExistingError:
            raise
        except Exception as e:
            raise mEx.ChartError("Error setting title for X Axis of chart document.") from e

    @staticmethod
    def set_y_axis_title(chart_doc: XChartDocument, title: str) -> None:
        """
        Sets the Y Axis Title

        Args:
            chart_doc (XChartDocument): Chart Document
            title (str): Title text

        Raises:
            DiagramNotExistingError: If ``chart_doc`` diagram does not exist
            ServiceNotSupported: If ``com.sun.star.chart.ChartAxisYSupplier`` service is not supported by diagram.
            ChartError: If any other error occurs.

        Returns:
            None:
        """
        try:
            diagram = chart_doc.getDiagram()
            if diagram is None:
                raise mEx.DiagramNotExistingError("Diagram does not exist")

            if not mInfo.Info.support_service(diagram, "com.sun.star.chart.ChartAxisYSupplier"):
                raise mEx.ServiceNotSupported("com.sun.star.chart.ChartAxisYSupplier")

            mProps.Props.set(diagram, HasYAxisDescription=True)
            axis = mLo.Lo.qi(XAxisYSupplier, diagram, True)
            title_shape = axis.getYAxisTitle()
            mProps.Props.set(title_shape, String=title)
            # mProps.Props.set(title_shape, HasYAxisTitle=True)
        except mEx.DiagramNotExistingError:
            raise
        except Exception as e:
            raise mEx.ChartError("Error setting title for Y Axis of chart document.") from e

    @staticmethod
    def set_second_y_axis(chart_doc: XChartDocument, title: str) -> None:
        """
        Sets the second Y Axis Title

        Args:
            chart_doc (XChartDocument): Chart Document
            title (str): Title text

        Raises:
            DiagramNotExistingError: If ``chart_doc`` diagram does not exist
            ServiceNotSupported: If ``com.sun.star.chart.ChartTwoAxisYSupplier`` service is not supported by diagram.
            ChartError: If any other error occurs.

        Returns:
            None:
        """
        try:
            diagram = chart_doc.getDiagram()
            if diagram is None:
                raise mEx.DiagramNotExistingError("Diagram does not exist")

            if not mInfo.Info.support_service(diagram, "com.sun.star.chart.ChartTwoAxisYSupplier"):
                raise mEx.ServiceNotSupported("com.sun.star.chart.ChartTwoAxisYSupplier")

            mProps.Props.set(diagram, HasSecondaryYAxis=True)
            axis = mLo.Lo.qi(XTwoAxisYSupplier, diagram, True)

            # y2_ps = axis.getSecondaryYAxis()
            # mProps.Props.show_props("Second y-axis", y2_ps)

            y2_title_ps = mLo.Lo.qi(XPropertySet, axis.getYAxisTitle(), True)
            mProps.Props.set(y2_title_ps, String=title)
            # mProps.Props.show_props("Second y-axis title", y2_title_ps)

        except mEx.DiagramNotExistingError:
            raise
        except Exception as e:
            raise mEx.ChartError("Error setting second Y Axis of chart document.") from e

    @staticmethod
    def view_legend(chart_doc: XChartDocument, is_visible: bool) -> None:
        """
        Set the visibility of chart document legend

        Args:
            chart_doc (XChartDocument): Chart Document
            is_visible (bool): Visibility

        Raises:
            ServiceNotSupported: If ``com.sun.star.chart.ChartDocument`` service is not supported.
            ChartError: If any other error occurs.

        Returns:
            None:
        """
        # this was showLegend in java
        if not mInfo.Info.support_service(chart_doc, "com.sun.star.chart.ChartDocument"):
            raise mEx.ServiceNotSupported("com.sun.star.chart.ChartDocument")

        try:

            mProps.Props.set_property(chart_doc, "HasLegend", is_visible)
        except Exception as e:
            raise mEx.ChartError(f'Error setting chart document legend visibility to "{is_visible}"') from e

    @staticmethod
    def set_use_horizontals(chart_doc: XChartDocument, is_vertical: bool = True) -> None:
        """
        Determines if the bars of a Bar Diagram chart are drawn vertically or horizontally.
        Default is vertical.

        If ``is_vertical`` is ``False`` you get a column chart rather than a bar chart.

        Args:
            chart_doc (XChartDocument): Bar Diagram chart document
            is_vertical (bool, optional): Determines if chart is drawn vertically or horizontally. Default ``True``.

        Raises:
            DiagramNotExistingError: If ``chart_doc`` diagram does not exist
            ServiceNotSupported: If ``com.sun.star.chart.BarDiagram`` service is not supported by diagram.
            ChartError: If any other error occurs.

        Returns:
            None:
        """
        # was useHorizontals in java
        try:
            diagram = chart_doc.getDiagram()
            if diagram is None:
                raise mEx.DiagramNotExistingError("Diagram does not exist")

            if not mInfo.Info.support_service(diagram, "com.sun.star.chart.BarDiagram"):
                raise mEx.ServiceNotSupported("com.sun.star.chart.BarDiagram")

            mProps.Props.set_property(diagram, "Vertical", is_vertical)

        except mEx.DiagramNotExistingError:
            raise
        except Exception as e:
            raise mEx.ChartError("Error setting use horizontals of chart document.") from e

    @staticmethod
    def set_use_lines(chart_doc: XChartDocument, has_lines: bool) -> None:
        """
        Sets if the chart type has lines connecting the data points or contains just symbols.

        Args:
            chart_doc (XChartDocument): ``LineDiagram`` or ``XYDiagram`` chart document.
            has_lines (bool): Determines if the chart type has lines connecting the data points or contains just symbols.

        Raises:
            DiagramNotExistingError: If ``chart_doc`` diagram does not exist
            ServiceNotSupported: If chart document is not a ``LineDiagram`` or ``XYDiagram``
            ChartError: If any other error occurs.

        Returns:
            None:
        """
        # usesLines in java
        try:
            # com.sun.star.chart.XYDiagram inherits from LineDiagram
            diagram = chart_doc.getDiagram()
            if diagram is None:
                raise mEx.DiagramNotExistingError("Diagram does not exist")

            if not mInfo.Info.support_service(
                diagram, "com.sun.star.chart.LineDiagram", "com.sun.star.chart.XYDiagram"
            ):
                raise mEx.ServiceNotSupported("com.sun.star.chart.LineDiagram", "com.sun.star.chart.XYDiagram")

            mProps.Props.set(diagram, Lines=has_lines)

        except mEx.ServiceNotSupported:
            raise
        except mEx.DiagramNotExistingError:
            raise
        except Exception as e:
            raise mEx.ChartError("Error setting use lines chart document.") from e

    @staticmethod
    def set_data_caption(chart_doc: XChartDocument, label_types: ChartDataCaptionEnum) -> None:
        """
        Sets how the caption of data points is displayed.

        Args:
            chart_doc (XChartDocument): Chart Document
            label_types (ChartDataCaptionEnum): Flags, specifies how the caption of data points is displayed.

        Raises:
            DiagramNotExistingError: If ``chart_doc`` diagram does not exist
            ServiceNotSupported: If ``com.sun.star.chart.Diagram`` service is not supported by diagram.
            ChartError: If any other error occurs.

        Returns:
            None:
        """
        # ChartDataCaptionEnum is IntFlags enum.
        try:
            diagram = chart_doc.getDiagram()
            if diagram is None:
                raise mEx.DiagramNotExistingError("Diagram does not exist")

            if not mInfo.Info.support_service(diagram, "com.sun.star.chart.Diagram"):
                raise mEx.ServiceNotSupported("com.sun.star.chart.Diagram")

            mProps.Props.set(diagram, DataCaption=int(label_types))

        except mEx.ServiceNotSupported:
            raise
        except mEx.DiagramNotExistingError:
            raise
        except Exception as e:
            raise mEx.ChartError("Error setting data caption of chart document.") from e

    @staticmethod
    def set_data_placement(chart_doc: XChartDocument, placement: DataLabelPlacementEnum) -> None:
        """
        Sets the relative position for the data label.

        Args:
            chart_doc (XChartDocument): Chart Document.
            placement (DataLabelPlacementEnum): Specifies a relative position for the data label.

        Raises:
            DiagramNotExistingError: If ``chart_doc`` diagram does not exist
            ServiceNotSupported: If ``com.sun.star.chart.ChartDataPointProperties`` service is not supported by diagram.
            ChartError: If any other error occurs.

        Returns:
            None:
        """
        try:
            diagram = chart_doc.getDiagram()
            if diagram is None:
                raise mEx.DiagramNotExistingError("Diagram does not exist")

            if not mInfo.Info.support_service(diagram, "com.sun.star.chart.ChartDataPointProperties"):
                raise mEx.ServiceNotSupported("com.sun.star.chart.ChartDataPointProperties")

            data_arr = mLo.Lo.qi(XChartDataArray, chart_doc.getData(), True)
            adata = data_arr.getData()
            num_points = len(adata)
            for i in range(num_points):
                # first parameter is the index of the point, the second one is the series
                point_props = diagram.getDataPointProperties(i, 0)

                # point_props.setPropertyValue("CharHeight", 14.0)
                # point_props.setPropertyValue("CharWeight", FontWeight.BOLD)
                # point_props.setPropertyValue("CharColor", 0x993366)
                point_props.setPropertyValue("LabelPlacement", int(placement))

            if mInfo.Info.support_service(diagram, "com.sun.star.chart.PieDiagram"):
                # for pie chart
                ps = diagram.getDataPointProperties(0, 0)
            else:
                ps = diagram.getDataRowProperties(0)
            mProps.Props.set(ps, LabelPlacement=int(placement))
            mProps.Props.show_props("Data Row", ps)

        except mEx.ServiceNotSupported:
            raise
        except mEx.DiagramNotExistingError:
            raise
        except Exception as e:
            raise mEx.ChartError("Error setting data placement of chart document.") from e

    @staticmethod
    def show_data_array(chart_doc: XChartDocument) -> None:
        """
        Prints data array info to console.

        Args:
            chart_doc (XChartDocument): Chart Document.
        """
        chart_data = chart_doc.getData()
        data_arr = mLo.Lo.qi(XChartDataArray, chart_data, True)
        data = data_arr.getData()
        if not data:
            print("No data found")
        else:
            print(f"No. of Data columns: {len(data[0])}")
            print(f"No. of Data rows: {len(data)}")
            for row in data:
                for col in row:
                    print(f"  {col}")
        print()
        row_descs = data_arr.getRowDescriptions()
        if not row_descs:
            print("No row description found")
        else:
            print(f"No. of rows: {len(row_descs)}")
            for row in row_descs:
                print(f'  "{row}"')
        print()

        col_descs = data_arr.getColumnDescriptions()
        if not col_descs:
            print("No column description found")
        else:
            print(f"No. of columns: {len(row_descs)}")
            for col in col_descs:
                print(f'  "{col}"')
        print()

    @staticmethod
    def set_area_transparency(chart_doc: XChartDocument, val: Intensity) -> None:
        """
        Set the transparency of chart areas surrounding the diagram/

        Args:
            chart_doc (XChartDocument): Chart Document.
            val (Intensity): Transparency intensity, Higher values is more transparent.

        Raises:
            ChartError: If error occurs.

        Returns:
            None:
        """
        try:
            mProps.Props.set(chart_doc.getArea(), FillTransparence=val.Value)
        except Exception as e:
            raise mEx.ChartError("Error setting chart transparency") from e

    # region set_3d()
    @overload
    @staticmethod
    def set_3d(chart_doc: XChartDocument, is_3d: bool) -> None:
        ...

    @overload
    @staticmethod
    def set_3d(chart_doc: XChartDocument, is_3d: bool, solid_type: ChartSolidTypeEnum) -> None:
        ...

    @staticmethod
    def set_3d(chart_doc: XChartDocument, is_3d: bool, solid_type: ChartSolidTypeEnum | None = None) -> None:
        """
        Sets chart 3d option.

        Args:
            chart_doc (XChartDocument): Chart Document.
            is_3d (bool): Specifies if chart is 3d.
            solid_type (ChartSolidTypeEnum): Chart Solid Type. Defaults to ``ChartSolidTypeEnum.CYLINDER``.

        Raises:
            DiagramNotExistingError: If ``chart_doc`` diagram does not exist
            ChartError: If any other error occurs.

        Returns:
            None:
        """
        try:
            diagram = chart_doc.getDiagram()
            if diagram is None:
                raise mEx.DiagramNotExistingError("Diagram does not exist")

            is_vert = bool(mProps.Props.get(diagram, "Vertical"))
            if solid_type is None:
                solid_type = ChartSolidTypeEnum.CYLINDER
            mProps.Props.set(diagram, Dim3D=is_3d, SolidType=int(solid_type), Vertical=is_vert)

        except mEx.DiagramNotExistingError:
            raise
        except Exception as e:
            raise mEx.ChartError("Error setting chart to 3d.") from e

    # endregion set_3d()

    @staticmethod
    def set_symbol(chart_doc: XChartDocument, has_symbol: bool) -> None:
        """
        Sets chart symbol.

        Args:
            chart_doc (XChartDocument): Chart Document.
            has_symbol (bool): Specifies if chart is 3d.

        Raises:
            DiagramNotExistingError: If ``chart_doc`` diagram does not exist
            ServiceNotSupported: If chart document is not a ``LineDiagram`` or ``XYDiagram``
            ChartError: If any other error occurs.

        Returns:
            None:
        """
        # In this interface, only the two values ChartSymbolType::NONE and
        # ChartSymbolType::AUTO are supported. Later versions may support the selection of the symbols shape.

        # XYDiagram inherits LineDiagram
        try:
            diagram = chart_doc.getDiagram()
            if diagram is None:
                raise mEx.DiagramNotExistingError("Diagram does not exist")

            if not mInfo.Info.support_service(
                diagram, "com.sun.star.chart.LineDiagram", "com.sun.star.chart.XYDiagram"
            ):
                raise mEx.ServiceNotSupported("com.sun.star.chart.LineDiagram", "com.sun.star.chart.XYDiagram")

            if has_symbol:
                mProps.Props.set(diagram, SymbolType=ChartSymbolType.AUTO)
            else:
                mProps.Props.set(diagram, SymbolType=ChartSymbolType.NONE)

        except mEx.ServiceNotSupported:
            raise
        except mEx.DiagramNotExistingError:
            raise
        except Exception as e:
            raise mEx.ChartError("Error setting chart symbol.") from e

    @staticmethod
    def set_trend(chart_doc: XChartDocument, curve_type: ChartRegressionCurveType) -> None:
        """
        Sets chart trend

        Args:
            chart_doc (XChartDocument): Chart Document.
            curve_type (ChartRegressionCurveType): determines a type of regression for the data row values.

        Raises:
            DiagramNotExistingError: If ``chart_doc`` diagram does not exist
            ServiceNotSupported: If ``com.sun.star.chart.ChartStatistics`` service is not supported by diagram.
            ChartError: If any other error occurs.

        Returns:
            None:
        """
        # determines a type of regression for the data row values.
        try:
            diagram = chart_doc.getDiagram()
            if diagram is None:
                raise mEx.DiagramNotExistingError("Diagram does not exist")

            if not mInfo.Info.support_service(diagram, "com.sun.star.chart.ChartStatistics"):
                raise mEx.ServiceNotSupported("com.sun.star.chart.ChartStatistics")

            # data_arr = mLo.Lo.qi(XChartDataArray, chart_doc.getData(), True)
            # data = data_arr.getData()
            # num_points = len(data)

            if (
                curve_type == ChartRegressionCurveType.NONE
                or curve_type == ChartRegressionCurveType.LINEAR
                or curve_type == ChartRegressionCurveType.LOGARITHM
                or curve_type == ChartRegressionCurveType.EXPONENTIAL
                or curve_type == ChartRegressionCurveType.POWER
            ):
                mProps.Props.set(diagram, RegressionCurves=curve_type)
            else:
                mLo.Lo.print(f"Did not recognize curve type: {curve_type}")

        except mEx.ServiceNotSupported:
            raise
        except mEx.DiagramNotExistingError:
            raise
        except Exception as e:
            raise mEx.ChartError("Error setting chart trend.") from e

    @staticmethod
    def set_data_row_descriptions(chart_doc: XChartDocument, *descs: str) -> None:
        """
        Set data row descriptions for chart.

        Args:
            chart_doc (XChartDocument): Chart Document.
            *descs (str): Variable length argument list of descriptions. Must match the same number of descriptions in the chart.

        Raises:
            ValueError: If wrong number of descriptions are provided.
            ChartError: If any other error occurs

        Returns:
            None:

        Note:
            This method works, but affects the graphic data point position.
        """
        try:
            chart_data = chart_doc.getData()
            data_arr = mLo.Lo.qi(XChartDataArray, chart_data, True)

            row_descs = data_arr.getRowDescriptions()
            if not row_descs:
                mLo.Lo.print("No row description found")
                return

            if len(row_descs) != len(descs):
                raise ValueError(f"Row length mismatch; No. of rows == {len(row_descs)}")
            data_arr.setRowDescriptions(descs)
        except ValueError:
            raise
        except Exception as e:
            raise mEx.ChartError("Error setting chart row descriptions.") from e

    @staticmethod
    def show_data_row_props(chart_doc: XChartDocument) -> None:
        """
        Prints data row info to console.

        Args:
            chart_doc (XChartDocument): Chart Document.
        """
        try:
            diagram = chart_doc.getDiagram()
            if diagram is None:
                raise mEx.DiagramNotExistingError("Diagram does not exist")

            data_arr = mLo.Lo.qi(XChartDataArray, chart_doc.getData(), True)
            data = data_arr.getData()
            num_points = len(data)
            print(f"No. of points: {num_points}")

            for i in range(num_points):
                point_ps = diagram.getDataPointProperties(i, 0)
                mProps.Props.show_props(f"{i}. Data Point", point_ps)

            if mInfo.Info.support_service(diagram, "com.sun.star.chart.PieDiagram"):
                # for pie chart
                ps = diagram.getDataPointProperties(0, 0)
            else:
                ps = diagram.getDataRowProperties(0)

            mProps.Props.show_props("Data Row", ps)
        except Exception as e:
            print("Could not get DataPointProperties:")
            print(f"  {e}")

    # endregion adjust properties

    # endregion background colors


__all__ = ("Chart",)



class GlobalViewDispatch:
    """
    Global View Dispatch Commands

    See Also:
        - :py:meth:`.Lo.dispatch_cmd`
        - `Global Dispatch commands <https://wiki.documentfoundation.org/Development/DispatchCommands#Global>`_
    """

    ACTIVATE = "Activate"
    BEAMER = "Beamer"
    BROWSE_VIEW = "BrowseView"
    """Web"""
    COMMAND_POPUP = "CommandPopup"
    """Search Commands"""
    CONTEXT = "Context"
    """Current Context"""
    DESIGNER_DIALOG = "DesignerDialog"
    """Show the Styles Sidebar"""
    FULL_SCREEN = "FullScreen"
    """Full Screen"""
    GRID_USE = "GridUse"
    """Snap to Grid"""
    GRID_VISIBLE = "GridVisible"
    """Display Grid"""
    HELPLINES_MOVE = "HelplinesMove"
    INSPECT_SELECTED_OBJECT = "InspectSelectedObject"
    IS_LOADING = "IsLoading"
    IS_LOADING_IMAGES = "IsLoadingImages"
    MODIFIED = "Modified"
    MODIFIED_STATUS = "ModifiedStatus"
    NEW_WINDOW = "NewWindow"
    """New Window"""
    NOTEBOOKBAR = "Notebookbar"
    OUTLINE_COLLAPSE = "OutlineCollapse"
    """Hide Sub-points"""
    OUTLINE_COLLAPSE_ALL = "OutlineCollapseAll"
    """Show Only First Level"""
    OUTLINE_DOWN = "OutlineDown"
    """Move selected paragraphs down one paragraph"""
    OUTLINE_EXPAND = "OutlineExpand"
    """Show Sub-points"""
    OUTLINE_EXPAND_ALL = "OutlineExpandAll"
    """Show All Levels"""
    OUTLINE_LEFT = "OutlineLeft"
    """Promote outline level of selected list paragraphs"""
    OUTLINE_RIGHT = "OutlineRight"
    """Demote outline level of selected list paragraphs"""
    OUTLINE_UP = "OutlineUp"
    """Move selected paragraphs up one paragraph"""
    PRINT_PREVIEW = "PrintPreview"
    """Close Preview"""
    REFRESH_VIEW = "RefreshView"
    """Refresh document layout"""
    RESTORE_EDITING_VIEW = "RestoreEditingView"
    """Restore Editing View"""
    SHOW_ANNOTATIONS = "ShowAnnotations"
    """Comments"""
    SIDEBAR = "Sidebar"
    SIDEBAR_MENU = "SidebarMenu"
    SIZE = "Size"
    SM_EDIT_OPTIONS = "SmEditOptions"
    """Formula Options"""
    STATE_TABLE_CELL = "StateTableCell"
    """Cell"""
    STATUS_BAR_VISIBLE = "StatusBarVisible"
    """Status Bar"""
    SWITCH_VIEW_SHELL = "SwitchViewShell"
    ZOOM = "Zoom"
    ZOOM_100_PERCENT = "Zoom100Percent"
    """100%"""
    ZOOM_150_PERCENT = "Zoom150Percent"
    """150%"""
    ZOOM_200_PERCENT = "Zoom200Percent"
    """200%"""
    ZOOM_50_PERCENT = "Zoom50Percent"
    """50%"""
    ZOOM_75_PERCENT = "Zoom75Percent"
    """75%"""
    ZOOM_MENU = "ZoomMenu"
    ZOOM_MINUS = "ZoomMinus"
    """Zoom Out"""
    ZOOM_NEXT = "ZoomNext"
    """Zoom Next"""
    ZOOM_OBJECTS = "ZoomObjects"
    """Object Zoom"""
    ZOOM_PAGE = "ZoomPage"
    """Entire Page"""
    ZOOM_PAGE_WIDTH = "ZoomPageWidth"
    """Page Width"""
    ZOOM_PLUS = "ZoomPlus"
    """Zoom In"""
    ZOOM_PREVIOUS = "ZoomPrevious"
    """Zoom Previous"""
    ZOOM_TOOL_BOX = "ZoomToolBox"
    """Zoom"""



# coding: utf-8
from __future__ import annotations
import uno
import unohelper
from typing import Tuple
from com.sun.star.datatransfer import XTransferable
from ooo.dyn.datatransfer.unsupported_flavor_exception import UnsupportedFlavorException
from ooo.dyn.datatransfer.data_flavor import DataFlavor


class ImageTransferable(unohelper.Base, XTransferable):

    MIME_TYPE = 'application/x-openoffice-bitmap;windows_formatname="Bitmap"'

    def __init__(self) -> None:
        super().__init__()
        self._im_bytes = []

    def getTransferData(self, aFlavor: DataFlavor) -> object:
        """
        Called by a data consumer to obtain data from the source in a specified format.

        Args:
            aFlavor (DataFlavor): DataFlavor struct

        Raises:
            UnsupportedFlavorException: ``UnsupportedFlavorException``
            com.sun.star.io.IOException: ``IOException``
        """
        if aFlavor.MimeType == ImageTransferable.MIME_TYPE:
            return self._im_bytes
        raise UnsupportedFlavorException

    def getTransferDataFlavors(self) -> Tuple[DataFlavor, ...]:
        """
        Returns a sequence of supported DataFlavor.
        """
        # not yet sure the proper DataType.
        # In java the DataType is Type(byte[].class)
        dfs = (DataFlavor(MimeType=ImageTransferable.MIME_TYPE, HumanPresentableName="Bitmap", DataType=tuple),)
        return dfs

    def isDataFlavorSupported(self, aFlavor: DataFlavor) -> bool:
        """
        Checks if the data object supports the specified by :py:attr:`.ImageTransferable.MIME_TYPE`.

        A value of False if the DataFlavor is unsupported by the transfer source.

        Args:
            aFlavor (DataFlavor): DataFlavor struct

        Note:
            This method is only for analogy with the JAVA Clipboard interface.
            To avoid many calls, the caller should instead use
            ``com.sun.star.datatransfer.XTransferable.getTransferDataFlavors()``.
        """
        return aFlavor.MimeType == ImageTransferable.MIME_TYPE



# coding: utf-8
# Python conversion of XML.java by Andrew Davison, ad@fivedots.coe.psu.ac.th
# See Also: https://fivedots.coe.psu.ac.th/~ad/jlop/

# region Imports
from __future__ import annotations
from typing import TYPE_CHECKING
import os
from typing import Sequence, Tuple, List, overload
from xml.dom.minidom import Node, parse, Document, parseString
import urllib.request
from xml.dom.minicompat import NodeList
from ..exceptions import ex as mEx
from .table_helper import TableHelper
from . import lo as mLo  # lazy loading
from . import file_io as mFileIO
from .type_var import PathOrStr

# endregion Imports


class XML:
    """XML method used for with LibreOffice Documents"""

    # region --------------- Load / Save ------------------------------

    @classmethod
    def load_doc(cls, fnm: PathOrStr) -> Document:
        """
        Gets a document from a file

        Args:
            fnm (PathOrStr): XML file to load.

        Raises:
            Exception: if unable to open document.

        Returns:
            Document: XML Document.
        """
        try:
            pth = mFileIO.FileIO.get_absolute_path(fnm)
            with open(pth) as file:
                doc = parse(file)
            cls._remove_whitespace(doc)
            doc.normalize()
            return doc
        except Exception as e:
            print(e)
            raise Exception(f"Opening of document failed: '{fnm}'") from e

    @classmethod
    def url_2_doc(cls, url: str) -> Document:
        """
        Gets a XML Document from remote source.

        Args:
            url (str): URL for a remote XML Document

        Raises:
            Exception: if unable to open document.

        Returns:
            Document: XML Document
        """
        try:
            with urllib.request.urlopen(url) as url_data:
                doc = parseString(url_data.read().decode())
            cls._remove_whitespace(doc)
            doc.normalize()
            return doc
        except Exception as e:
            print(e)
            raise Exception(f"Opening of document failed: '{url}'") from e

    @classmethod
    def str_to_doc(cls, xml_str: str) -> Document:
        """
        Gets a XML document from xml string.

        Args:
            xml_str (str): XML string.

        Raises:
            Exception: if unable to create document from xml.

        Returns:
            Document: XML Document on successful load; Otherwise, None.
        """
        try:
            doc = parseString(xml_str)
            cls._remove_whitespace(doc)
            doc.normalize()
            return doc
        except Exception as e:
            print(e)
            raise Exception(f"Error get xml docoument from xml string") from e

    @staticmethod
    def save_doc(doc: Document, xml_fnm: PathOrStr) -> None:
        """
        Save doc to xml file.

        Args:
            doc (Document): doc to save.
            xml_fnm (PathOrStr): Output file path.

        Raises:
            Exception: If unable to save document
        """
        try:
            pth = mFileIO.FileIO.get_absolute_path(xml_fnm)
            with open(pth, "w") as file:

                sx: str = doc.toprettyxml(indent="  ")
                # remove any empty lines, there if often a lot with toprettyxml()
                lines = [line for line in sx.splitlines() if line.strip() != ""]
                lines.append("")  # end with empty line
                clean_sx = "\n".join(lines)
                file.write(clean_sx)
                # doc.writexml(writer=file, indent="  ")
        except Exception as e:
            raise Exception(f"Unable to save document to {xml_fnm}") from e

    # endregion ------------ Load / Save ------------------------------

    # region --------------- DOM data extraction -----------------------

    @staticmethod
    def get_node(tag_name: str, nodes: NodeList) -> Node | None:
        """
        Gets the fist tag_name found in nodes.

        Args:
            tag_name (str): tag name to find in nodes.
            nodes (NodeList): Nodes to search

        Returns:
            Node | None: First found node; Otherwise, None
        """
        name = tag_name.casefold()
        for node in nodes:
            if node.nodeType == Node.ELEMENT_NODE and node.tagName.casefold() == name:
                return node
        return None

    # region    get_node_value()
    @overload
    @staticmethod
    def get_node_value(node: Node) -> str:
        """
        Get the text stored in the node

        Args:
            node (Node): Node to get value of.

        Returns:
            str: Node value.
        """
        ...

    @overload
    @staticmethod
    def get_node_value(tag_name: str, nodes: NodeList) -> str:
        """
        Gets firt tag_name node in the list and returns it text.

        Args:
            tag_name (str): tag_name to search for.
            nodes (NodeList): List of nodes to search.

        Returns:
            str: Node value if found; Otherwise empty str.
        """
        ...

    @classmethod
    def get_node_value(cls, *args, **kwargs) -> str:
        """
        Gets first ``tag_name`` node in the list and returns it text.

        Args:
            node (Node): Node to get value of.
            tag_name (str): ``tag_name`` to search for.
            nodes (NodeList): List of nodes to search.

        Returns:
            str: Node value if found; Otherwise empty str.
        """
        ordered_keys = (1, 2)
        kargs_len = len(kwargs)
        count = len(args) + kargs_len

        def get_kwargs() -> dict:
            ka = {}
            if kargs_len == 0:
                return ka
            valid_keys = ("tag_name", "nodes", "node")
            check = all(key in valid_keys for key in kwargs.keys())
            if not check:
                raise TypeError("get_node_value() got an unexpected keyword argument")
            keys = ("tag_name", "node")
            for key in keys:
                if key in kwargs:
                    ka[1] = kwargs[key]
                    break
            if count == 1:
                return ka
            ka[2] = kwargs.get("nodes", None)
            return ka

        if not count in (1, 2):
            raise TypeError("get_node_value() got an invalid numer of arguments")

        kargs = get_kwargs()

        for i, arg in enumerate(args):
            kargs[ordered_keys[i]] = arg

        if count == 1:
            return cls._get_node_val(kargs[1])
        return cls._get_node_val2(kargs[1], kargs[2])

    @staticmethod
    def _get_node_val(node: Node) -> str:
        if node is None:
            return ""
        if not node.hasChildNodes():
            return ""
        child_nodes: NodeList = node.childNodes
        for node in child_nodes:
            if node.nodeType == Node.TEXT_NODE:
                return str(node.data).strip()
        return ""

    @classmethod
    def _get_node_val2(cls, tag_name: str, nodes: NodeList) -> str:
        if nodes is None:
            return ""
        name = tag_name.casefold()
        for node in nodes:
            if node.nodeName.casefold() == name:
                return cls._get_node_val(node)
        return ""

    # endregion get_node_value()

    @classmethod
    def get_node_values(cls, nodes: NodeList) -> Tuple[str, ...]:
        """
        Gets all the node values

        Args:
            nodes (NodeList): Nodes to get values of.

        Returns:
            Tuple[str, ...]: Node Values
        """
        vals = []
        for node in nodes:
            val = cls._get_node_val(node)
            if val != "":
                vals.append(val)
        return tuple(val)

    @staticmethod
    def get_node_attr(attr_name: str, node: Node) -> str:
        """
        Get the named attribute value from node

        Args:
            attr_name (str): Attribute Name
            node (Node): Node to get attribute of.

        Returns:
            str: Attribute value if found; Otherwise empty str.
        """
        if node.attributes is None:
            return ""
        # attrs is {} if there are no attributes
        attrs = dict(node.attributes.items())
        name = attr_name.casefold()
        for k, v in attrs.items():
            if str(k).casefold() == name:
                return str(v)
        return ""

    @classmethod
    def get_all_node_values(cls, row_nodes: NodeList, col_ids: Sequence[str]) -> List[list] | None:
        """
        Gets all node values.

        .. collapse:: Example XML

            XML is assumed to have structure that is similar

                .. include:: ../../resources/xml/pay.xml.rst

        The data from a sequence of <col> becomes one row in the
        generated 2D array.

        The first row of the 2D array contains the col ID strings.

        Args:
            row_nodes (NodeList): rows
            col_ids (Sequence[str]): Column ids

        Returns:
            List[list] | None: 2D-list of values on success; Otherwise, None

        Note:
            col_ids must match the column names:

            ``colids = ("purpose", "amount", "tax", "maturity")``

            Results for example xml:

            .. include:: ../../resources/xml/pay_all_notes_result.rst
        """
        num_rows = len(row_nodes)
        num_cols = len(col_ids)
        if num_cols == 0 or num_rows == 0:
            return None
        data = TableHelper.make_2d_array(num_rows=num_rows, num_cols=num_cols)
        # data = [[1] * num_cols for _ in range(num_rows + 1)]
        # put column strings in first row of list
        for col, _ in enumerate(col_ids):
            data[0][col] = mLo.Lo.capitalize(col_ids[col])

        for i, node in enumerate(row_nodes):
            # extract all the column strings for ith row
            col_nodes = node.childNodes
            for col in range(num_cols):
                data[i][col] = cls.get_node_value(col_ids[col], col_nodes)
        return data

    # endregion ------------ DOM data extraction -----------------------

    # region ---------------- XLS transforming -------------------------

    @staticmethod
    def apply_xslt(xml_fnm: PathOrStr, xls_fnm: PathOrStr) -> str:
        """
        Transforms xml file using XLST.

        Not available in macros at this time.

        Args:
            xml_fnm (PathOrStr): XML source file path.
            xls_fnm (PathOrStr): XSL source file path.

        Raises:
            NotSupportedMacroModeError: If access in a macro
            Exception: If lxml python package is not available
            Exception: If unable to apply xls

        Returns:
            str: String of XML that has been transformed.
        """
        if mLo.Lo.is_macro_mode:
            raise mEx.NotSupportedMacroModeError("apply_xslt() is not supported from a macro")
        try:
            from lxml import etree as XML_ETREE
        except ImportError as e:
            raise Exception("apply_xslt() requires lxml python package") from e
        _xml_parser = XML_ETREE.XMLParser(remove_blank_text=True)

        try:
            pth_xml = mFileIO.FileIO.get_absolute_path(xml_fnm)
            pth_xls = mFileIO.FileIO.get_absolute_path(xls_fnm)
            print(f"Applying filter '{xls_fnm}' to '{xml_fnm}'")
            dom = XML_ETREE.parse(pth_xml, parser=_xml_parser)
            xslt = XML_ETREE.parse(pth_xls)
            transform = XML_ETREE.XSLT(xslt)
            newdom = transform(dom)
            t_result = XML_ETREE.tostring(newdom, encoding="unicode")  # unicode produces string
            return t_result
        except Exception as e:
            raise Exception(f"Unable to transform '{xml_fnm}' with '{xls_fnm}'") from e

    @staticmethod
    def apply_xslt_to_str(xml_str: str, xls_fnm: PathOrStr) -> str:
        """
        Transforms xml using XLST.

        Not available in macros at this time.

        Args:
            xml_str (str): Raw XML data.
            xls_fnm (PathOrStr): XSL source file path.

        Raises:
            NotSupportedMacroModeError: If access in a macro
            Exception: If lxml python package is not available
            Exception: If unable to apply xls

        Returns:
            str: String of XML that has been transformed.
        """
        if mLo.Lo.is_macro_mode:
            raise mEx.NotSupportedMacroModeError("apply_xslt_2_str() is not supported from a macro")
        try:
            from lxml import etree as XML_ETREE
        except ImportError as e:
            raise Exception("apply_xslt requires lxml python package") from e
        _xml_parser = XML_ETREE.XMLParser(remove_blank_text=True)

        try:
            pth = mFileIO.FileIO.get_absolute_path(xls_fnm)
            print(f"Applying the filter in '{xls_fnm}'")
            dom = XML_ETREE.fromstring(xml_str)
            xslt = XML_ETREE.parse(pth, parser=_xml_parser)

            transform = XML_ETREE.XSLT(xslt)
            newdom = transform(dom)
            t_result = XML_ETREE.tostring(newdom, encoding="unicode")  # unicode produces string
            return t_result
        except Exception as e:
            raise Exception("Unable to transform the string") from e

    # endregion ------------- XLS transforming -------------------------

    # region --------------- Filter ------------------------------------

    @staticmethod
    def get_flat_fiter_name(doc_type: mLo.Lo.DocTypeStr) -> str:
        """
        Gets the Flat XML filter name for the doc type.

        Args:
            doc_type (Lo.DocTypeStr): Document type.

        Returns:
            str: Flat XML filter name.
        """
        if doc_type == mLo.Lo.DocTypeStr.WRITER:
            return "OpenDocument Text Flat XML"
        elif doc_type == mLo.Lo.DocTypeStr.CALC:
            return "OpenDocument Spreadsheet Flat XML"
        elif doc_type == mLo.Lo.DocTypeStr.DRAW:
            return "OpenDocument Drawing Flat XML"
        elif doc_type == mLo.Lo.DocTypeStr.IMPRESS:
            return "OpenDocument Presentation Flat XML"
        else:
            print("No Flat XML filter for this document type; using Flat text")
            return "OpenDocument Text Flat XML"

    # endregion ------------ Filter ------------------------------------

    # region --------------- Formating --------------------------------

    # region    indent()
    @overload
    @classmethod
    def indent(cls, src: str) -> str:
        """
        Indents xml

        Args:
            src (str): raw xml data.

        Raises:
            TypeError is src is not expected type
            Exception: If unable to indent

        Returns:
            str: Indented xml as string.
        """
        ...

    @overload
    @classmethod
    def indent(cls, src: os.PathLike) -> str:
        """
        Indents xml

        Args:
            src (PathLike): xml file path.

        Raises:
            TypeError is src is not expected type
            Exception: If unable to indent

        Returns:
            str: Indented xml as string.
        """
        ...

    @overload
    @classmethod
    def indent(cls, src: Document) -> str:
        """
        Indents xml

        Args:
            src (Document): xml doucment.

        Raises:
            TypeError is src is not expected type
            Exception: If unable to indent

        Returns:
            str: Indented xml as string.
        """
        ...

    @classmethod
    def indent(cls, src: os.PathLike | str | Document) -> str:
        """
        Indents xml

        Args:
            src (str | PathLike | Document): raw xml data or xml file path or xml document.

        Raises:
            TypeError is src is not expected type
            Exception: If unable to indent

        Returns:
            str: Indented xml as string.
        """
        try:
            if isinstance(src, os.PathLike):
                with open(mFileIO.FileIO.get_absolute_path(src), "r") as file:
                    doc = parse(file)
            elif isinstance(src, str):
                doc = parseString(src)
            elif isinstance(src, Document):
                # don't modify origin document
                doc = parseString(src.toxml())
            else:
                raise TypeError(
                    f"src is not recognized. Expected, str, PathLike or Document. Got {type(src).__name__}"
                )
            cls._remove_whitespace(doc)
            doc.normalize()
            # To parse string instead use: dom = md.parseString(xml_string)
            pretty_xml = doc.toprettyxml()
            # remove the weird newline issue:
            # should not be needes with cls._remove_whitespace(doc)
            # pretty_xml = os.linesep.join([s for s in pretty_xml.splitlines() if s.strip()])
            return pretty_xml
        except TypeError:
            raise
        except Exception as e:
            if isinstance(src, (str, os.PathLike)):
                msg = f"Unable to indent '{src}'"
            else:
                msg = f"Unable to indent document"
            raise Exception(msg) from e

    # endregion indent()

    @classmethod
    def _remove_whitespace(cls, node):
        """
        Removes whites from xml node

        Args:
            node (node): xml node, or xml document

        Note:
            it is necessary .normalize() the document to combine adjacent text nodes.
            Otherwise, you could end up with a bunch of redundant XML elements with just whitespace.
            Again, recursion is the only way to visit tree elements since you can’t iterate over the
            document and its elements with a loop. Finally, this should give you the expected result:
        """
        # https://realpython.com/python-xml-parser/
        # e.g.
        # document = parse("smiley.svg")
        # cls._remove_whitespace(document)
        # document.normalize()
        if node.nodeType == Node.TEXT_NODE:
            if node.nodeValue.strip() == "":
                node.nodeValue = ""
        for child in node.childNodes:
            cls._remove_whitespace(child)

    # @classmethod
    # def _indent(cls, elem, level=0) -> None:
    #     # pretty print
    #     # https://stackoverflow.com/questions/749796/pretty-printing-xml-in-python
    #     i = "\n" + level*"  "
    #     if len(elem):
    #         if not elem.text or not elem.text.strip():
    #             elem.text = i + "  "
    #         if not elem.tail or not elem.tail.strip():
    #             elem.tail = i
    #         for elem in elem:
    #             cls._indent(elem, level+1)
    #         if not elem.tail or not elem.tail.strip():
    #             elem.tail = i
    #     else:
    #         if level and (not elem.tail or not elem.tail.strip()):
    #             elem.tail = i

    # endregion ------------- Formating --------------------------------



from enum import IntEnum, unique
from . import kind_helper


@unique
class PresentationLayoutKind(IntEnum):
    """
    ``com.sun.star.presentation.DrawPage`` service ``Layout`` property values.

    See Also:
        `API Presentation Drawpage Service <https://api.libreoffice.org/docs/idl/ref/servicecom_1_1sun_1_1star_1_1presentation_1_1DrawPage.html>`_

    Example:
        .. code-block:: python

            Props.set(slide, Layout=PresentationLayoutKind.TITLE_ONLY.value)
    """

    TITLE_SUB = 0
    TITLE_BULLETS = 1
    TITLE_CHART = 2
    TITLE_2CONTENT = 3
    TITLE_CONTENT_CHART = 4
    TITLE_CONTENT_CLIP = 6
    TITLE_CHART_CONTENT = 7
    TITLE_TABLE = 8
    TITLE_CLIP_CONTENT = 9
    TITLE_CONTENT_OBJECT = 10
    TITLE_OBJECT = 11
    TITLE_CONTENT_2CONTENT = 12
    TITLE_OBJECT_CONTENT = 13
    TITLE_CONTENT_OVER_CONTENT = 14
    TITLE_2CONTENT_CONTENT = 15
    TITLE_2CONTENT_OVER_CONTENT = 16
    TITLE_CONTENT_OVER_OBJECT = 17
    TITLE_4OBJECT = 18
    TITLE_ONLY = 19
    BLANK = 20
    VTITLE_VTEXT_CHART = 27
    VTITLE_VTEXT = 28
    TITLE_VTEXT = 29
    TITLE_VTEXT_CLIP = 30
    CENTERED_TEXT = 32
    TITLE_4CONTENT = 33
    TITLE_6CONTENT = 34

    @staticmethod
    def from_str(s: str) -> "PresentationLayoutKind":
        """
        Gets an ``PresentationLayoutKind`` instance from string.

        Args:
            s (str): String that represents the name of an enum Attribute or enum value as string.
                ``s`` is case insensitive and can be ``CamelCase``, ``pascal_case`` , ``snake_case``,
                ``hypen-case``, ``normal case``.

        Raises:
            ValueError: If input string is empty.
            AttributeError: If unable to get ``PresentationLayoutKind`` instance.

        Returns:
            PresentationLayoutKind: Enum instance.
        """
        return kind_helper.enum_from_string(s, PresentationLayoutKind)



class GlobalFormatDispatch:
    """
    Global Format Dispatch Commands

    See Also:
        - :py:meth:`.Lo.dispatch_cmd`
        - `Global Dispatch commands <https://wiki.documentfoundation.org/Development/DispatchCommands#Global>`_
    """

    ALIGN_CENTER = "AlignCenter"
    """Centered"""
    ALIGN_DOWN = "AlignDown"
    """Bottom"""
    ALIGN_MIDDLE = "AlignMiddle"
    """Center"""
    ALIGN_UP = "AlignUp"
    """Top"""
    AUTO_FORMAT = "AutoFormat"
    """
    Auto-Format Styles

    Args
        - ``aFormatName (str)``
    """
    BACKGROUND_COLOR = "BackgroundColor"
    """
    Background Color

    Args
        - ``Color (str)``
        - ``BackgroundColor  (color)``
    """
    BACKGROUND_PATTERN_CONTROLLER = "BackgroundPatternController"
    """Background Pattern"""
    BOLD = "Bold"
    """Bold"""
    BOLD_LATIN = "BoldLatin"
    BRING_TO_FRONT = "BringToFront"
    """Bring to Front"""
    CENTER_PARA = "CenterPara"
    """Align Center"""
    CHANGE_CASE_ROTATE_CASE = "ChangeCaseRotateCase"
    """
    Cycle Case ``(Title Case, Sentence case, UPPERCASE, lowercase)``
    """
    CHANGE_CASE_TO_FULL_WIDTH = "ChangeCaseToFullWidth"
    """Full-width"""
    CHANGE_CASE_TO_HALF_WIDTH = "ChangeCaseToHalfWidth"
    """Half-width"""
    CHANGE_CASE_TO_HIRAGANA = "ChangeCaseToHiragana"
    """Hiragana"""
    CHANGE_CASE_TO_KATAKANA = "ChangeCaseToKatakana"
    """Katakana"""
    CHANGE_CASE_TO_LOWER = "ChangeCaseToLower"
    """``lowercase``"""
    CHANGE_CASE_TO_SENTENCE_CASE = "ChangeCaseToSentenceCase"
    """Sentence case"""
    CHANGE_CASE_TO_TITLE_CASE = "ChangeCaseToTitleCase"
    """Capitalize Every Word"""
    CHANGE_CASE_TO_TOGGLE_CASE = "ChangeCaseToToggleCase"
    """``tOGGLE cASE``"""
    CHANGE_CASE_TO_UPPER = "ChangeCaseToUpper"
    """``UPPERCASE``"""
    CHAR_BACK_COLOR = "CharBackColor"
    """
    Args
        - ``Color (str)``
        - ``CharBackColor (color)``
    """
    CHAR_FONT_NAME = "CHAR_FONT_NAME"
    """FONT NAME"""
    CHAR_FONT_NAME_LATIN = "CharFontNameLatin"
    CHARACTER_BACKGROUND_PATTERN = "CharacterBackgroundPattern"
    """Highlight Color"""
    COLOR = "Color"
    """
    Font Color

    Args
        - ``Color (str)``
        - ``Color (color)``
        - ``ColorThemeIndex (int)``
        - ``ColorLumMod (int)``
        - ``ColorLumOff (int)``
    """
    COMMON_ALIGN_BOTTOM = "CommonAlignBottom"
    """Bottom"""
    COMMON_ALIGN_HORIZONTAL_CENTER = "CommonAlignHorizontalCenter"
    """Centered"""
    COMMON_ALIGN_HORIZONTAL_DEFAULT = "CommonAlignHorizontalDefault"
    """Default"""
    COMMON_ALIGN_JUSTIFIED = "CommonAlignJustified"
    """Justified"""
    COMMON_ALIGN_LEFT = "CommonAlignLeft"
    """Left"""
    COMMON_ALIGN_RIGHT = "CommonAlignRight"
    """Right"""
    COMMON_ALIGN_TOP = "CommonAlignTop"
    """Top"""
    COMMON_ALIGN_VERTICAL_CENTER = "CommonAlignVerticalCenter"
    """Center"""
    COMMON_ALIGN_VERTICAL_DEFAULT = "CommonAlignVerticalDefault"
    COMMON_TASK_BAR_VISIBLE = "CommonTaskBarVisible"
    """Presentation"""
    """Default"""
    DECREMENT_INDENT = "DecrementIndent"
    """Decrease Indent"""
    EMPHASIS_MARK = "EmphasisMark"
    ENTER_GROUP = "EnterGroup"
    """Enter Group"""
    FILL_COLOR = "FillColor"
    """
    Fill Color

    Args
        - ``Color (str)``
        - ``FillColor (xfillcolor)``
        - ``ColorThemeIndex (int)``
        - ``ColorLumMod (int)``
        - ``ColorLumOff (int)``
    """
    FILL_FLOAT_TRANSPARENCE = "FillFloatTransparence"
    """Gradient Fill Transparency"""
    FILL_STYLE = "FillStyle"
    """Area Style / Filling"""
    FILL_TRANSPARENCE = "FillTransparence"
    """Fill Transparency"""
    FILTER_CRIT = "FilterCrit"
    """Standard Filter"""
    FIND_ALL = "FindAll"
    """Find All"""
    FIND_TEXT = "FindText"
    """Find text in values, to search in formulas use the dialog"""
    FIT_CELL_SIZE = "FitCellSize"
    "Fit to Cell Size"
    FLIP_HORIZONTAL = "FlipHorizontal"
    """Flip Horizontally"""
    FLIP_VERTICAL = "Flip Vertically"
    FONT_HEIGH_LATIN = "FontHeighLatin"
    FONT_HEIGHT = "FontHeight"
    """Font Size"""
    FONT_WORK = "FontWork"
    FORMAT_AREA = "FormatArea"
    FORMAT_BULLETS_MENU = "FormatBulletsMenu"
    FORMAT_FORM_MENU = "FormatFormMenu"
    FORMAT_FRAME_MENU = "FormatFrameMenu"
    FORMAT_GROUP = "FormatGroup"
    """Group"""
    FORMAT_IMAGE_FILTERS_MENU = "FormatImageFiltersMenu"
    """Filter"""
    FORMAT_IMAGE_MENU = "FormatImageMenu"
    """Image"""
    FORMAT_LINE = "FormatLine"
    """Line"""
    FORMAT_SPACING_MENU = "FormatSpacingMenu"
    """Spacing"""
    FORMAT_STYLES_MENU = "FormatStylesMenu"
    """Styles"""
    FORMAT_TEXT_MENU = "FormatTextMenu"
    """Text"""
    FORMAT_UNGROUP = "FormatUngroup"
    """``Un-group``"""
    GRAF_BLUE = "GrafBlue"
    """Blue"""
    GRAF_CONTRAST = "GrafContrast"
    """Contrast"""
    GRAF_GAMMA = "GrafGamma"
    """Gamma"""
    GRAF_GREEN = "GrafGreen"
    """Green"""
    GRAF_INVERT = "GrafInvert"
    """Invert"""
    GRAF_LUMINANCE = "GrafLuminance"
    """Brightness"""
    GRAF_MODE = "GrafMode"
    """Image Mode"""
    GRAF_RED = "GrafRed"
    """Red"""
    GRAF_TRANSPARENCE = "GrafTransparence"
    """Transparency"""
    ITALIC = "Italic"
    ITALIC_LATIN = "ItalicLatin"
    """Italic Latin"""
    JUSTIFY_PARA = "JustifyPara"
    """Justified"""
    LANGUAGE_LATIN = "LanguageLatin"
    LANGUAGE_STATUS = "LanguageStatus"
    """Language Status"""
    LEAVEGROUP = "LeaveGroup"
    """Exit Group"""
    LEFT_PARA = "LeftPara"
    """Align Left"""
    LINE_CAP = "LineCap"
    LINE_DASH = "LineDash"
    LINE_END_STYLE = "LineEndStyle"
    """
    Select start and end arrowheads for lines.

    Args
        - ``LineStart (xlinestart)``
        - ``LineEnd (xlineend)``
        - ``StartWidth (int)``
        - ``EndWidth (int)``
    """
    LINE_JOINT = "LineJoint"
    """Line Corner Style"""
    LINE_SPACING = "LineSpacing"
    """Set Line Spacing"""
    LINE_TRANSPARENCE = "LineTransparence"
    """Line Transparency"""
    LINE_WIDTH = "LineWidth"
    """
    Line Width

    Args
        - ``Width (float)``
        - ``LineWidth (LineWidth)``
    """
    MEASURE_ATTRIBUTES = "MeasureAttributes"
    """Dimensions"""
    OUTLINE_FONT = "OutlineFont"
    """Apply outline attribute to font. Not all fonts implement this attribute."""
    OUTLINE_FORMAT = "OutlineFormat"
    """Show Formatting"""
    OVERLINE = "Overline"
    PARA_LEFT_TO_RIGHT = "ParaLeftToRight"
    """Left-To-Right"""
    PARA_RIGHT_TO_LEFT = "ParaRightToLeft"
    """Right-To-Left"""
    PARAGRAPH_DIALOG = "ParagraphDialog"
    """Paragraph"""
    PARASPACE_DECREASE = "ParaspaceDecrease"
    "Decrease Paragraph Spacing"
    PARASPACE_INCREASE = "ParaspaceIncrease"
    """Increase Paragraph Spacing"""
    RIGHT_PARA = "RightPara"
    """Align Right"""
    RUBY_DIALOG = "RubyDialog"
    """Asian Phonetic Guide"""
    RULER_MENU = "RulerMenu"
    """Rulers"""
    SEND_TO_BACK = "SendToBack"
    """Send to Back"""
    SET_DEFAULT = "SetDefault"
    """Clear Direct Formatting"""
    SET_OBJECT_TO_BACKGROUND = "SetObjectToBackground"
    """To Background"""
    SET_OBJECT_TO_FOREGROUND = "SetObjectToForeground"
    """To Foreground"""
    SHADOWED = "Shadowed"
    """Toggle Shadow"""
    SHAPES_LINE_MENU = "ShapesLineMenu"
    """Line"""
    SHAPES_MENU = "ShapesMenu"
    """Shape"""
    SHRINK = "Shrink"
    """Decrease Font Size"""
    SMALL_CAPS = "SmallCaps"
    """Small capitals"""
    SPACE_PARA_1 = "SpacePara1"
    """Line Spacing: 1"""
    SPACE_PARA_1_15 = "SpacePara115"
    """Line Spacing: 1.15"""
    SPACE_PARA_1_5 = "SpacePara15"
    """Line Spacing: 1.5"""
    SPACE_PARA_2 = "SpacePara2"
    "Line Spacing: 2"
    SPACING = "Spacing"
    """Set Character Spacing"""
    SPELL_CHECK_APPLY_SUGGESTION = "SpellCheckApplySuggestion"
    """
    Apply Suggestion

    Args
        - ``ApplyRule (str)``
    """
    SPELL_CHECK_IGNORE = "SpellCheckIgnore"
    """Ignore"""
    SPELL_CHECK_IGNORE_ALL = "SpellCheckIgnoreAll"
    """Ignore All"""
    STRIKE_OUT = "Strikeout"
    """Strike through"""
    TABLE_DESIGN = "TableDesign"
    """Table Design"""
    TABLE_STYLE_SETTINGS = "TableStyleSettings"
    """
    Args
        - ``UseFirstRowStyle (bool)``
        - ``UseLastRowStyle (bool)``
        - ``UseBandingRowStyle (bool)``
        - ``UseFirstColumnStyle (bool)``
        - ``UseLastColumnStyle (bool)``
        - ``UseBandingColumnStyle (bool)``
    """
    TEXT_FIT_TO_SIZE = "TextFitToSize"
    """Fit to Frame"""
    TEXT_DIRECTION_TOP_TO_BOTTOM = "TextdirectionTopToBottom"
    """Text direction from top to bottom"""
    TRANSFORM_DIALOG = "TransformDialog"
    """
    Position and Size

    Args
        - `TransformPosX (int)``
        - ``TransformPosY (int)``
        - ``TransformWidth (int)``
        - ``TransformHeight (int)``
        - ``TransformRotationDeltaAngle (sdrangle)``
        - ``TransformRotationAngle (sdrangle)``
        - `TransformRotationX (int)``
        - ``TransformRotationY (int)``
    """
    TRANSFORM_ROTATION_ANGLE = "TransformRotationAngle"
    """Rotation Angle"""
    TRANSFORM_ROTATION_X = "TransformRotationX"
    """Rotation Pivot Point X"""
    TRANSFORM_ROTATION_Y = "TransformRotationY"
    """Rotation Pivot Point Y"""
    UNDERLINE = "Underline"
    UNDERLINE_DOTTED = "UnderlineDotted"
    """Dotted Underline"""
    UNDERLINE_DOUBLE = "UnderlineDouble"
    """Double Underline"""
    UNDERLINE_NONE = "UnderlineNone"
    """Underline: Off"""
    UNDERLINE_SINGLE = "UnderlineSingle"
    """Single Underline"""
    X_LINE_COLOR = "XLineColor"
    """
    Line Color

    Args
        - `Color (str)``
        - ``XLineColor (xlinecolor)``
    """
    X_LINE_STYLE = "XLineStyle"
    """Line Style"""



# coding: utf-8
from __future__ import annotations
from typing import Any
from abc import ABC


class AbstractEvent(ABC):
    # https://stackoverflow.com/questions/472000/usage-of-slots
    __slots__ = ()

    def __init__(self, source: Any) -> None:
        """
        Constructor

        Args:
            source (Any): Event Source
        """
        self.source = source
        self._event_name = ""
        self._event_source = None
        self.event_data = None

    source: Any
    """Gets/Sets Event source"""
    event_data: Any
    """Gets/Sets any extra data associated with the event"""

    @property
    def event_name(self) -> str:
        """
        Gets the event name for these args
        """
        return self._event_name

    @property
    def event_source(self) -> Any | None:
        """
        Gets the event source for these args
        """
        return self._event_source

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}: {self.event_name}>"


class EventArgs(AbstractEvent):
    """
    Event Arguments Class
    """

    __slots__ = ("source", "_event_name", "event_data", "_event_source")

    @staticmethod
    def from_args(args: EventArgs) -> EventArgs:
        """
        Gets a new instance from existing instance

        Args:
            args (EventArgs): Existing Instance

        Returns:
            EventArgs: args
        """
        eargs = EventArgs(source=args.source)
        eargs._event_name = args.event_name
        eargs._event_source = args.event_source
        eargs.event_data = args.event_data
        return eargs


e = EventArgs(None)



# PM - Notes, Oct 25, 2022
#
# this module has issues.
# find_gallery_item() returns XGalleryItem which in may have properties such as Graphic (XGraphic).
# the issues is as soon as the XGalleryItem is returned from find_gallery_item() it has already lost its properties.
# see my post: https://ask.libreoffice.org/t/libreoffice-wiping-object-properties-issue-with-gallerythemeprovider/83182
#
# Checking the reference count inside of find_gallery_item() show there is only 1 ref could. Which means 0 references
# becuase sys.getrefcount() add 1 reference to the object that is being checked.
# find_gallery_graphic() has been added which duplicates the code of find_gallery_item() due to these issues.
# find_gallery_graphic() does successfully return XGraphic object, However I am not sure if this graphic can be used
# as it is. I tried inserting the XGraphic into a Draw XShape and putting it on the document, however it appears to
# always be the same graphic even though the criteria is change for find_gallery_graphic()

from __future__ import annotations
from typing import TYPE_CHECKING, Any
from pathlib import Path
from typing import overload
import uno
from com.sun.star.gallery import XGalleryItem
from com.sun.star.gallery import XGalleryTheme
from com.sun.star.gallery import XGalleryThemeProvider
from com.sun.star.graphic import XGraphic

from . import file_io as mFileIo
from . import lo as mLo
from . import props as mProps
from . import info as mInfo
from ..exceptions import ex as mEx
from .kind.search_match_kind import SearchMatchKind as SearchMatchKind
from .kind.gallery_kind import GalleryKind as GalleryKind
from .kind.gallery_search_by_kind import SearchByKind as SearchByKind

from ..events.event_singleton import _Events
from ..events.lo_named_event import LoNamedEvent

from ooo.dyn.gallery.gallery_item_type import GalleryItemTypeEnum as GalleryItemTypeEnum
from ooo.lo.gallery.gallery_item_type import GalleryItemType as GalleryItemType

from ..meta.static_meta import StaticProperty, classproperty

if TYPE_CHECKING:
    from com.sun.star.lang import EventObject


class Gallery(metaclass=StaticProperty):
    # region get_gallery()
    @overload
    @staticmethod
    def get_gallery(name: GalleryKind) -> XGalleryTheme:
        ...

    @overload
    @staticmethod
    def get_gallery(name: str) -> XGalleryTheme:
        ...

    @staticmethod
    def get_gallery(name: GalleryKind | str) -> XGalleryTheme:

        # enusre name is GalleryKind | str,
        mInfo.Info.is_type_enum_multi(alt_type="str", enum_type=GalleryKind, enum_val=name, arg_name="name")

        try:
            gtp = mLo.Lo.create_instance_mcf(
                XGalleryThemeProvider, "com.sun.star.gallery.GalleryThemeProvider", raise_err=True
            )
            obj = gtp.getByName(str(name))
            if obj is None:
                raise mEx.UnKnownError("None Value: getByName() returned None value")
            return mLo.Lo.qi(XGalleryTheme, obj, True)
        except Exception as e:
            raise mEx.GalleryError(f'Error occured getting gallery for "{name}"') from e

    # endregion get_gallery()

    @staticmethod
    def get_item_type_str(item_type: GalleryItemTypeEnum) -> str:
        if item_type == GalleryItemTypeEnum.EMPTY:
            return "empty"
        elif item_type == GalleryItemTypeEnum.GRAPHIC:
            return "graphic"
        elif item_type == GalleryItemTypeEnum.MEDIA:
            return "media"
        elif item_type == GalleryItemType.DRAWING:
            return "drawing"
        return "??"  # just in case enum change in future.

    @staticmethod
    def get_item_type(item: XGalleryItem) -> GalleryItemTypeEnum:
        try:
            item_int_type = int(mProps.Props.get(item, "GalleryItemType"))
            return GalleryItemTypeEnum(item_int_type)
        except Exception as e:
            raise mEx.GalleryError("Error getting item type from gallery item") from e

    @classmethod
    def get_gallery_graphic(cls, item: XGalleryItem) -> XGraphic:
        """
        Gets graphic form item

        Args:
            item (XGalleryItem): Gallery Item

        Raises:
            GalleryTypeError: If ``item.Graphic`` is not ``GalleryItemType.DRAWING``
            GalleryError: If any other error occurs.

        Returns:
            XGraphic: Graphic object
        """
        try:
            if item is None:
                raise TypeError("item is expected to be of Type XGalleryItem, got None")
            itm_type = cls.get_item_type(item)
            if itm_type != GalleryItemTypeEnum.DRAWING:
                raise mEx.GalleryTypeError(
                    f" Expected item to be GalleryItemType.DRAWING but got GalleryItemType.{itm_type.name}"
                )
            result = mProps.Props.get(item, "Graphic")
            return mLo.Lo.qi(XGraphic, result, True)
        except mEx.GalleryTypeError:
            raise
        except Exception as e:
            raise mEx.GalleryError("Error getting gallery graphic") from e

    @staticmethod
    def get_gallery_path(item: XGalleryItem) -> Path:
        if item is None:
            raise TypeError("item is expected to be of Type XGalleryItem, got None")
        try:
            url = str(mProps.Props.get(item, "URL"))
            return mFileIo.FileIO.url_to_path(url)
        except Exception as e:
            raise mEx.GalleryError("Error getting gallery path") from e

    # region find_gallery_item()
    @overload
    @classmethod
    def find_gallery_item(cls, gallery_name: str, name: str) -> XGalleryItem:
        ...

    @overload
    @classmethod
    def find_gallery_item(cls, gallery_name: GalleryKind, name: str) -> XGalleryItem:
        ...

    @overload
    @classmethod
    def find_gallery_item(cls, gallery_name: str, name: str, search_match: SearchMatchKind) -> XGalleryItem:
        ...

    @overload
    @classmethod
    def find_gallery_item(cls, gallery_name: GalleryKind, name: str, search_match: SearchMatchKind) -> XGalleryItem:
        ...

    @overload
    @classmethod
    def find_gallery_item(cls, gallery_name: str, name: str, *, search_kind: SearchByKind) -> XGalleryItem:
        ...

    @overload
    @classmethod
    def find_gallery_item(cls, gallery_name: GalleryKind, name: str, *, search_kind: SearchByKind) -> XGalleryItem:
        ...

    @overload
    @classmethod
    def find_gallery_item(
        cls, gallery_name: str, name: str, search_match: SearchMatchKind, search_kind: SearchByKind
    ) -> XGalleryItem:
        ...

    @overload
    @classmethod
    def find_gallery_item(
        cls, gallery_name: GalleryKind, name: str, search_match: SearchMatchKind, search_kind: SearchByKind
    ) -> XGalleryItem:
        ...

    @classmethod
    def find_gallery_item(
        cls,
        gallery_name: GalleryKind | str,
        name: str,
        search_match: SearchMatchKind = SearchMatchKind.PARTIAL_IGNORE_CASE,
        search_kind: SearchByKind = SearchByKind.TITLE,
    ) -> XGalleryItem:

        result = None
        try:
            if search_match == SearchMatchKind.FULL_IGNORE_CASE or search_kind == SearchMatchKind.PARTIAL_IGNORE_CASE:
                nm = name.casefold()
                case_sensitive = False
            else:
                nm = name
                case_sensitive = True

            if search_match == SearchMatchKind.PARTIAL or search_match == SearchMatchKind.PARTIAL_IGNORE_CASE:
                partial = True
            else:
                partial = False

            gallery = cls.get_gallery(gallery_name)
            num_pics = gallery.getCount()
            mLo.Lo.print(f'Searching gallery "{gallery.getName()}" for "{name}"')
            if case_sensitive:
                mLo.Lo.print("  Search is case sensitive")
            else:
                mLo.Lo.print("  Search is ignoring case")
            if partial:
                mLo.Lo.print("  Searching for a partial match")
            else:
                mLo.Lo.print("  Searching for a full match")
            mLo.Lo.print()

            for i in range(num_pics):
                item = mLo.Lo.qi(XGalleryItem, gallery.getByIndex(i), True)
                if search_kind == SearchByKind.FILE_NAME:
                    url = str(mProps.Props.get(item, "URL"))
                    fnm = mFileIo.FileIO.get_fnm((mFileIo.FileIO.url_to_path(url)))
                    match_str = fnm if case_sensitive else fnm.casefold()
                    if partial and match_str in nm:
                        mLo.Lo.print(f"Found matching item: {fnm}")
                        result = item
                        break
                    else:
                        if match_str == nm:
                            mLo.Lo.print(f"Found matching item: {fnm}")
                            result = item
                            break
                elif search_kind == SearchByKind.TITLE:
                    title = str(mProps.Props.get(item, "Title"))
                    match_str = title if case_sensitive else title.casefold()
                    if partial and match_str in nm:
                        mLo.Lo.print(f"Found matching item: {title}")
                        result = item
                        break
                    else:
                        if match_str == nm:
                            mLo.Lo.print(f"Found matching item: {title}")
                            result = item
                            break

        except Exception as e:
            raise mEx.GalleryError(
                f'Error occured trying to find in gallery: "{gallery_name}" for item: "{name}"'
            ) from e
        if result is None:
            raise mEx.GalleryNotFoundError(f'Not found. Tried to find in gallery: "{gallery_name}" for item: "{name}"')
        cls.report_gallery_item(result)
        return result

    # endregion find_gallery_item()

    # region find_gallery_graphic()
    @overload
    @classmethod
    def find_gallery_graphic(cls, name: str) -> XGraphic:
        ...

    @overload
    @classmethod
    def find_gallery_graphic(cls, name: str, search_match: SearchMatchKind) -> XGraphic:
        ...

    @overload
    @classmethod
    def find_gallery_graphic(cls, name: str, *, search_kind: SearchByKind) -> XGraphic:
        ...

    @overload
    @classmethod
    def find_gallery_graphic(cls, name: str, search_match: SearchMatchKind, search_kind: SearchByKind) -> XGraphic:
        ...

    @classmethod
    def find_gallery_graphic(
        cls,
        name: str,
        search_match: SearchMatchKind = SearchMatchKind.PARTIAL_IGNORE_CASE,
        search_kind: SearchByKind = SearchByKind.TITLE,
    ) -> XGraphic:

        result = None
        try:
            if search_match == SearchMatchKind.FULL_IGNORE_CASE or search_kind == SearchMatchKind.PARTIAL_IGNORE_CASE:
                nm = name.casefold()
                case_sensitive = False
            else:
                nm = name
                case_sensitive = True

            if search_match == SearchMatchKind.PARTIAL or search_match == SearchMatchKind.PARTIAL_IGNORE_CASE:
                partial = True
            else:
                partial = False

            gallery = cls.get_gallery(GalleryKind.SHAPES)
            num_pics = gallery.getCount()
            mLo.Lo.print(f'Searching gallery "{gallery.getName()}" for "{name}"')
            if case_sensitive:
                mLo.Lo.print("  Search is case sensitive")
            else:
                mLo.Lo.print("  Search is ignoring case")
            if partial:
                mLo.Lo.print("  Searching for a partial match")
            else:
                mLo.Lo.print("  Searching for a full match")
            mLo.Lo.print()

            for i in range(num_pics):
                item = mLo.Lo.qi(XGalleryItem, gallery.getByIndex(i), True)
                if search_kind == SearchByKind.FILE_NAME:
                    url = str(mProps.Props.get(item, "URL"))
                    fnm = mFileIo.FileIO.get_fnm((mFileIo.FileIO.url_to_path(url)))
                    match_str = fnm if case_sensitive else fnm.casefold()
                    if partial and match_str in nm:
                        mLo.Lo.print(f"Found matching item: {fnm}")
                        result = item
                        break
                    else:
                        if match_str == nm:
                            mLo.Lo.print(f"Found matching item: {fnm}")
                            result = item
                            break
                elif search_kind == SearchByKind.TITLE:
                    title = str(mProps.Props.get(item, "Title"))
                    match_str = title if case_sensitive else title.casefold()
                    if partial and match_str in nm:
                        mLo.Lo.print(f"Found matching item: {title}")
                        result = item
                        break
                    else:
                        if match_str == nm:
                            mLo.Lo.print(f"Found matching item: {title}")
                            result = item
                            break

        except Exception as e:
            raise mEx.GalleryError(
                f'Error occured trying to find in gallery: "{GalleryKind.SHAPES}" for item: "{name}"'
            ) from e
        # Gallery._TMP_FIND = result.Graphic
        if result is None:
            raise mEx.GalleryNotFoundError(
                f'Not found. Tried to find in gallery: "{GalleryKind.SHAPES}" for item: "{name}"'
            )
        try:
            graphic = mProps.Props.get(result, "Graphic")
        except mEx.PropertyNotFoundError as e:
            raise mEx.GalleryNotFoundError("Error getting Graphic Property") from e
        return graphic

    # endregion find_gallery_graphic()

    @classmethod
    def report_gallery_item(cls, item: XGalleryItem) -> None:
        if item is None:
            print("Gallery item is null")
            return

        print("Gallery item information:")
        try:
            url = str(mProps.Props.get(item, "URL"))
            path = mFileIo.FileIO.uri_to_path(uri_fnm=url, ensure_absolute=False)
            print(f'  URL: "{url}"')
            print(f'  Fnm: "{Path(Gallery.gallery_dir, mFileIo.FileIO.get_fnm(path))}"')
            print(f"  Path: {path}")
        except mEx.PropertyNotFoundError:
            print("  URL: Property NOT Found")
            print("  Fnm: Unable to compute due to no URL Property")
            print("  Path: Unable to compute due do no URL Property")
        try:
            print(f'  Title: "{mProps.Props.get(item, "Title")}"')
        except mEx.PropertyNotFoundError:
            print("  Title: Property NOT Found")
        try:
            item_int_type = int(mProps.Props.get(item, "GalleryItemType"))
            print(f"  Type: {cls.get_item_type_str(GalleryItemTypeEnum(item_int_type))}")
        except mEx.PropertyNotFoundError:
            print("  Type: Property NOT Found")

    @staticmethod
    def report_galeries() -> None:
        gtp = mLo.Lo.create_instance_mcf(
            XGalleryThemeProvider, "com.sun.star.gallery.GalleryThemeProvider", raise_err=True
        )
        themes = gtp.getElementNames()
        print(f"No. of themes: {len(themes)}")
        for theme in themes:
            try:
                gallery = mLo.Lo.qi(XGalleryTheme, gtp.getByName(theme), True)
                print(f'  "{gallery.getName()}" ({gallery.getCount()})')
            except Exception as e:
                print(f'Could not access gallery for "{theme}"')
                print(f"  {e}")
        print()

    @classmethod
    def report_gallery_items(cls, gallery_name: GalleryKind | str) -> None:
        try:
            gallery = cls.get_gallery(gallery_name)
            num_pics = gallery.getCount()
            print(f'Gallery: "{gallery.getName()}" ({num_pics})')
            for i in range(num_pics):
                try:
                    item = mLo.Lo.qi(XGalleryItem, gallery.getByIndex(i), True)
                    cls.report_gallery_item(item)
                    # url = str(mProps.Props.get(item, "URL"))
                    # print(f"  {mFileIo.FileIO.get_fnm((mFileIo.FileIO.url_to_path(url)))}")
                except Exception as ex:
                    print(f"Could not access galery item: {i}")
                    print(f"  {ex}")
        except Exception as e:
            print("Error reporting gallery items")
            print(f"  {e}")
        print()

    @classproperty
    def gallery_dir(cls) -> Path:
        """
        Get the first directory that contain the Gallery database and multimedia files.
        """
        try:
            return cls._gallery_dir
        except AttributeError:
            cls._gallery_dir = mInfo.Info.get_gallery_dir()
        return cls._gallery_dir


class _GalleryManager:
    """Manages clearing and resetting for Gallery static class"""

    @staticmethod
    def on_disposed(source: Any, event: EventObject) -> None:
        # Clean up static properties that may have been dynamically created.
        # print("Gallery Static Property Cleanup")
        dattrs = ("_gallery_dir",)
        for attr in dattrs:
            if hasattr(Gallery, attr):
                delattr(Gallery, attr)


_Events().on(LoNamedEvent.BRIDGE_DISPOSED, _GalleryManager.on_disposed)


__all__ = ("Gallery",)



from __future__ import annotations
import sys

try:
    from typing import Protocol
except ImportError:
    from typing_extensions import Protocol

from ..utils import type_var
from ..events.args import event_args


class EventObserver(Protocol):
    """
    Protocol Class for Event Observer.

    See Also:
        :py:mod:`~.events.lo_events`
    """

    def on(self, event_name: str, callback: type_var.EventCallback):
        """
        Registers an event

        Args:
            event_name (str): Unique event name
            callback (Callable[[object, EventArgs], None]): Callback function
        """
        ...

    def remove(self, event_name: str, callback: type_var.EventCallback) -> bool:
        """
        Removes an event callback

        Args:
            event_name (str): Unique event name
            callback (Callable[[object, EventArgs], None]): Callback function

        Returns:
            bool: True if callback has been removed; Otherwise, False.
            False means the callback was not found.
        """
        ...

    def trigger(self, event_name: str, event_args: event_args.EventArgs):
        """
        Trigger event(s) for a given name.

        Args:
            event_name (str): Name of event to trigger
            event_args (EventArgs): Event args passed to the callback for trigger.
        """
        ...
